//
// BrackRec.go - рекурсивный распознаватель вложенности круглых скобок.
// Построен на основе правил следующей q-грамматики:
//
// 1. S -> (B)B
// 2. B -> (B)B
// 3. B -> empty
//

package main

import (
    "fmt"
    "strings"
)

var str string      // Строка с входной цепочкой, имитирующая входную ленту
var i int           // Текущее положение входной головки
var brackets []rune // Срез, содержащий прочитанную строку со скобками

var erFlag int          // Флаг, фиксирующий наличие ошибок в середине правила

// Функция, реализующая чтение символов в буфер из входного потока.
// Используется для ввода с клавиатуры распознаваемой строки.
// Ввод осуществляется до нажатия на клавишу Enter.
// Символ '\n' яляется концевым маркером входной строки.
func GetOneLine() {
    fmt.Scanln(&str)
    str += "\n"; // Добавляется концевой маркер
}

// Функция, реализующая распознавание нетерминала B.
func B() bool {
    if(brackets[i] == '(') {
        i++
        if(B()) {
            if(brackets[i] == ')') {
                i++
                if(B()) {
                    // Если последний нетерминал корректен,
                    // то корректно и все правило.
                    return true
                } else { // Ошибка в нетерминале B
                    erFlag++
                    fmt.Print("Position ", i, ", ",
                              "Error 1: Incorrect last B!\n",
                        )
                    return false
                }
            } else { // Это не ')'
                erFlag++
                fmt.Print("Position ", i, ", ",
                          "Error 2: I want ')'!\n",
                    )
                return false
            }
        } else {
            erFlag++
            fmt.Print("Position ", i, ", ",
                      "Error 3: Incorrect first B!\n",
                )
            return false
        }
    } else {            // Смотрим другую альтернативу:
        return true     // Пустая цепочка допустима
    }
}

// Функция, реализующая распознавание нетерминала S.
func S() bool {
    if(brackets[i] == '(') {
        i++
        if(B()) {
            if(brackets[i] == ')') {
                i++
                if(B()) {
                    if(brackets[i] == '\n') {
                        return true     // за последним нетерминалом - конец
                    } else { // Там ничего не должно быть
                        erFlag++
                        fmt.Print("Position ", i, ", ",
                                  "Error 4: I want string end after last right bracket!\n",
                            )
                        return false
                    }
                } else {
                    erFlag++
                    fmt.Print("Position ", i, ", ",
                              "Error 4: Incorrect last B!\n",
                        )
                    return false
                }
            } else {
                erFlag++
                fmt.Print("Position ", i, ", ",
                          "Error 5: I want ')'!\n",
                    )
                return false
            }
        } else {
            erFlag++
            fmt.Print("Position ", i, ", ",
                      "Error 6: Incorrect first B!\n",
                )
            return false
        }
    }
    return false       // Первый символ цепочки некорректен
                       // что это ошибка, лучше определить снаружи
}

// Функция запускающая разбор и определяющая корректность его завершения,
// если первый символ не принадлежит цепочки
func Parser() bool {
    // Начальная инициализация.
    erFlag = 0
    i = 0;

    // Процесс пошел!
    if(S()) {
        return true     // Все прошло нормально
    } else {
        if(erFlag != 0) {
            fmt.Print("Position ", i, ", ",
                      "Error 7: Internal Error!\n",
                )
        } else {
            fmt.Print("Position ", i, ", ",
                      "Error 8: Incorrect first symbol of S!\n",
                )
        }
        return false    // Есть ошибки
    }
}

//=============================================================================
// Главная функция используется для тестирования до тех пор, 
// пока не будет прочитана пустая строка
func main() {
    // Цикл распознавания различных входных цепочек
    for {
        strCursor := ""
        str = "";
        
        // Чтение очередной входной цепочки в буфер
        fmt.Print("Input bracket's expression: ")

        // Формируем очередную строку скобок для распознавания.
        GetOneLine()
        brackets = []rune(str)

        // Здесь начинается разбор принятой строки.
        if Parser() {
            fmt.Print("+++++ OK! +++++\n")
        } else {
            fmt.Print("----- Fatal error (look upper error message)! -----\n")
        }
        // Вывод разобранной строки и значения позиции входой головки.
        fmt.Print("Line: ", str)
        strCursor = " Pos: " + strings.Repeat("-", i)
        strCursor +="^"
        fmt.Print(strCursor, "  i = ", i, "\n\n")

        if str == "\n" {
            break
        }
    }
    fmt.Println("Goodbye!")
}
